from pathlib import Path
from os.path import isfile
import json
from datetime import timedelta
from gi.repository import GObject
from gfeeds.conf_mapper import ConfMapper
from gfeeds.gsettings_wrapper import GsettingsWrapper
from gfeeds.signal_helper import signal_tuple
from gfeeds.util.paths import (
    ARTICLE_THUMB_CACHE_PATH,
    CACHE_HOME,
    CACHE_PATH,
    CONF_DIR,
    THUMBS_CACHE_PATH
)
from gfeeds.util.singleton import Singleton


class ConfManagerSignaler(GObject.Object):

    __gsignals__ = {
        'gfeeds_new_first_changed': signal_tuple(),
        'gfeeds_repopulation_required': signal_tuple(),
        'gfeeds_webview_settings_changed': signal_tuple(),
        'gfeeds_show_read_changed': signal_tuple(),
        'gfeeds_full_article_title_changed': signal_tuple(),
        'gfeeds_show_empty_feeds_changed': signal_tuple(),
        'gfeeds_full_feed_name_changed': signal_tuple(),
        'dark_mode_changed': signal_tuple(),
        'show_thumbnails_changed': signal_tuple(),
        'on_apply_adblock_changed': signal_tuple(),
        'on_refresh_blocklist': signal_tuple(),
        # Signals down here don't have to do with the config
        'gfeeds_filter_changed': signal_tuple(params=(GObject.TYPE_PYOBJECT,)),
        'gfeeds_tags_append': signal_tuple(params=(str,)),
        'gfeeds_tags_pop': signal_tuple(params=(str,)),
    }


def json_to_gsettings(gw: GsettingsWrapper, path: Path):
    conf = dict()

    if path.is_file():
        try:
            with open(path) as fd:
                conf = json.loads(fd.read())
        except Exception:
            return
    else:
        return

    for k in conf.keys():
        if k == 'windowsize':
            gw['window_width'] = conf[k]['width']
            gw['window_height'] = conf[k]['height']
            continue
        try:
            gw[k] = conf[k]
        except KeyError:
            print(f'json_to_gsettings: skipping unsupported key {k}')

    path.unlink()


class ConfManager(metaclass=Singleton):
    def __init__(self):
        self.signaler = ConfManagerSignaler()
        self.emit = self.signaler.emit
        self.connect = self.signaler.connect

        self.is_flatpak = isfile('/.flatpak-info')

        self.conf_dir = CONF_DIR
        self.cache_home = CACHE_HOME
        self.cache_path = CACHE_PATH
        self.thumbs_cache_path = THUMBS_CACHE_PATH
        for p in [
                self.conf_dir,
                self.cache_path,
                self.thumbs_cache_path,
        ]:
            if not p.is_dir():
                p.mkdir(parents=True)
        self.legacy_conf_path = self.conf_dir.joinpath(
            'org.gabmus.gfeeds.json'
        )

        self.gsettings_conf = GsettingsWrapper('org.gabmus.gfeeds')
        self.conf = self.gsettings_conf
        self.nconf = ConfMapper(self.gsettings_conf)
        json_to_gsettings(self.gsettings_conf, self.legacy_conf_path)

        self.article_thumb_cache_path = ARTICLE_THUMB_CACHE_PATH
        if not self.article_thumb_cache_path.is_file():
            self.article_thumb_cache = dict()
            self.save_article_thumb_cache()
        else:
            with open(self.article_thumb_cache_path, 'r') as fd:
                self.article_thumb_cache = json.loads(fd.read())

    def save_article_thumb_cache(self):
        with open(self.article_thumb_cache_path, 'w') as fd:
            fd.write(json.dumps(self.article_thumb_cache))

    @property
    def max_article_age(self) -> timedelta:
        return timedelta(days=self.nconf.max_article_age_days)

    def add_tag(self, tag: str, target_feeds=[]):
        tags = self.nconf.tags
        lowercase_tags = [t.lower() for t in tags]
        if tag.lower() not in lowercase_tags:
            tags.append(tag)
            self.nconf.tags = tags
            self.emit('gfeeds_tags_append', tag)
        feeds = self.nconf.feeds
        for feed in target_feeds:
            if 'tags' not in feeds[feed].keys():
                feeds[feed]['tags'] = []
            if tag not in feeds[feed]['tags']:
                feeds[feed]['tags'].append(tag)
        self.nconf.feeds = feeds

    def delete_tag(self, tag: str):
        tags = self.nconf.tags
        while tag in tags:
            tags.remove(tag)
        self.emit('gfeeds_tags_pop', tag)
        self.nconf.tags = tags
        self.remove_tag(tag, self.nconf.feeds.keys())

    def remove_tag(self, tag: str, target_feeds: list):
        feeds: dict = self.nconf.feeds
        for feed in target_feeds:
            if 'tags' not in feeds[feed].keys():
                continue
            if tag in feeds[feed]['tags']:
                feeds[feed].remove(tag)
        self.nconf.feeds = feeds

    # TODO: legacy; remove
    def save_conf(self, *_):
        pass



from typing import Optional
from gi.repository import GObject, Gtk, Adw
from gfeeds.feeds_manager import FeedsManager
from gfeeds.confManager import ConfManager


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/empty_state.ui')
class EmptyState(Adw.Bin):
    __gtype_name__ = 'EmptyState'

    def __init__(self):
        super().__init__()


class StackWithEmptyState(Gtk.Stack):
    __gtype_name__ = 'StackWithEmptyState'

    def __init__(self, main_widget: Optional[Gtk.Widget] = None):
        super().__init__(
            vexpand=True, hexpand=True,
            transition_type=Gtk.StackTransitionType.CROSSFADE
        )
        self.__main_widget = main_widget
        self.feedman = FeedsManager()
        self.confman = ConfManager()
        self.empty_state = EmptyState()
        if self.main_widget is not None:
            self.add_named(self.main_widget, 'main_widget')
        self.add_named(self.empty_state, 'empty_state')
        self.set_visible_child(
            self.main_widget if len(self.confman.nconf.feeds) > 0
            and self.main_widget is not None
            else self.empty_state
        )

        self.feedman.feed_store.connect(
            'items-changed',
            self.on_feed_store_items_changed
        )

    @GObject.Property(type=Gtk.Widget, default=None, nick='main-widget')
    def main_widget(self) -> Optional[Gtk.Widget]:
        return self.__main_widget

    @main_widget.setter
    def main_widget(self, w: Gtk.Widget):
        self.__main_widget = w
        self.add_named(self.__main_widget, 'main_widget')

    def on_feed_store_items_changed(self, *args):
        if len(self.feedman.feed_store) == 0:
            self.set_visible_child(self.empty_state)
        else:
            self.set_visible_child(self.main_widget)

    def on_feeds_append(self, *args):
        self.set_visible_child(self.main_widget)



from gettext import ngettext
from functools import reduce
from typing import Optional
from gfeeds.filter_view import FilterView
from gfeeds.feed_item import FeedItem
from gfeeds.stack_with_empty_state import StackWithEmptyState
from operator import or_
from subprocess import Popen
from gi.repository import GObject, Gtk, Adw, Gio
from gfeeds.sidebar import GFeedsSidebar
from gfeeds.webview import GFeedsWebView
from gfeeds.headerbar import LeftHeaderbar, RightHeaderbar
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/main_leaflet.ui')
class MainLeaflet(Adw.Bin):
    __gtype_name__ = 'MainLeaflet'
    left_box = Gtk.Template.Child()
    right_box = Gtk.Template.Child()
    leaflet = Gtk.Template.Child()
    connection_bar: Gtk.InfoBar = Gtk.Template.Child()
    left_headerbar: LeftHeaderbar = Gtk.Template.Child()
    filter_view: FilterView = Gtk.Template.Child()
    searchbar: Gtk.SearchBar = Gtk.Template.Child()
    searchbar_entry: Gtk.SearchEntry = Gtk.Template.Child()
    filter_flap: Adw.Flap = Gtk.Template.Child()
    sidebar_stack: StackWithEmptyState = Gtk.Template.Child()
    sidebar: GFeedsSidebar = Gtk.Template.Child()
    webview: GFeedsWebView = Gtk.Template.Child()
    right_headerbar: RightHeaderbar = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = FeedsManager()

        self.sidebar.listview_sw.connect_activate(
            self.on_sidebar_row_activated
        )
        self.filter_flap.bind_property(
            'reveal-flap', self.left_headerbar.filter_btn, 'active',
            GObject.BindingFlags.BIDIRECTIONAL
        )

        self.confman.connect(
            'gfeeds_filter_changed', self.on_filter_changed
        )
        self.searchbar_entry.connect(
            'changed',
            lambda entry: self.sidebar.set_search(entry.get_text())
        )
        self.searchbar.bind_property(
            'search-mode-enabled', self.left_headerbar.search_btn,
            'active', GObject.BindingFlags.BIDIRECTIONAL
        )

        self.feedman.connect(
            'feedmanager_refresh_end', self.on_refresh_end
        )
        self.feedman.connect(
            'feedmanager_online_changed',
            lambda _, value: self.connection_bar.set_revealed(not value)
        )

        self.on_leaflet_folded()

    def on_filter_changed(self, *_):
        self.left_headerbar.filter_btn.set_active(False)
        # reset vertical scroll position to 0
        adjustment = self.sidebar.listview_sw.get_vadjustment()
        adjustment.set_value(0)
        self.sidebar.listview_sw.set_vadjustment(adjustment)

    @Gtk.Template.Callback()
    def on_leaflet_folded(self, *_):
        rh = self.right_headerbar.right_headerbar
        lh = self.left_headerbar.left_headerbar
        if self.leaflet.get_folded():
            self.right_headerbar.back_btn.set_visible(True)
            rh.set_show_start_title_buttons(True)
            rh.set_show_end_title_buttons(True)
            lh.set_show_start_title_buttons(True)
            lh.set_show_end_title_buttons(True)
        else:
            self.right_headerbar.back_btn.set_visible(False)
            rh.set_show_start_title_buttons(False)
            rh.set_show_end_title_buttons(True)
            lh.set_show_start_title_buttons(True)
            lh.set_show_end_title_buttons(False)

    @Gtk.Template.Callback()
    def on_back_btn_clicked(self, *_):
        self.leaflet.set_visible_child(self.left_box)
        self.on_leaflet_folded()

    def on_view_mode_change(self, target):
        self.right_headerbar.on_view_mode_change(target)
        self.webview.change_view_mode(target)

    def on_refresh_end(self, *_):
        self.left_headerbar.errors_btn.set_visible(
            len(self.feedman.errors) > 0
        )
        self.sidebar.listview_sw.all_items_changed()
        self.sidebar.loading_revealer.set_running(False)
        if (
                self.confman.nconf.notify_new_articles and
                not self.get_root().is_active() and  # window is not focused
                self.feedman.new_items_num > 0
        ):
            notif_text = ngettext(
                '{0} new article', '{0} new articles',
                self.feedman.new_items_num
            ).format(self.feedman.new_items_num)
            notif = Gio.Notification.new(notif_text)
            notif.set_icon(Gio.ThemedIcon.new(
                'org.gabmus.gfeeds-symbolic'
            ))
            self.get_root().app.send_notification('new_articles', notif)

    def on_sidebar_row_activated(self, feed_item: Optional[FeedItem]):
        self.feedman.article_store.set_selected_article(feed_item)
        if not feed_item:
            return
        feed_item.read = True
        if (
                self.confman.nconf.open_youtube_externally and
                reduce(or_, [
                    f'://{pfx}' in feed_item.link  # type: ignore
                    for pfx in [
                        p + 'youtube.com'
                        for p in ('', 'www.', 'm.')
                    ]
                ])
        ):
            cmd_parts = [
                self.confman.nconf.media_player, f'"{feed_item.link}"'
            ]
            if self.confman.is_flatpak:
                cmd_parts.insert(0, 'flatpak-spawn --host')
            cmd = ' '.join(cmd_parts)
            Popen(cmd, shell=True)
            return
        self.webview.load_feeditem(feed_item)
        self.right_headerbar.set_article_title(
            feed_item.title
        )
        self.right_headerbar.extra_menu_btn.set_sensitive(True)
        self.leaflet.set_visible_child(self.right_box)
        self.on_leaflet_folded()
        self.sidebar.listview_sw.invalidate_filter()
        feed_item.emit('changed', '')



from gettext import gettext as _
from os.path import isfile
from pathlib import Path
import requests
from gfeeds.confManager import ConfManager
from gfeeds.util.create_full_url import create_full_url
from gfeeds.util.paths import CACHE_PATH
from gfeeds.util.sha import shasum
from syndom import Html
from typing import Literal, Optional, Union
from gfeeds.util.to_unicode import to_unicode, bytes_to_unicode

GET_HEADERS = {
    'User-Agent': 'gfeeds/1.0',
    'Accept': '*/*',
    'Accept-Encoding': 'gzip, deflate'
}

TIMEOUT = 30


class DownloadError(Exception):
    def __init__(self, code, *args):
        self.download_error_code = code


# will return the content of a file if it's a file url
def download_text(link: str) -> str:
    if link[:8] == 'file:///':
        with open(link[7:]) as fd:
            toret = fd.read()
        return toret
    res = requests.get(link, headers=GET_HEADERS, timeout=TIMEOUT)
    if 200 <= res.status_code <= 299:
        return bytes_to_unicode(res.content, enc=res.encoding)
    else:
        raise DownloadError(
            res.status_code, f'response code {res.status_code}'
        )


def download_raw(link: str, dest: str) -> None:
    res = requests.get(link, headers=GET_HEADERS, timeout=TIMEOUT)
    if res.status_code == 200:
        with open(dest, 'wb') as fd:
            for chunk in res.iter_content(1024):
                fd.write(chunk)
    else:
        raise requests.HTTPError(
            f'response code {res.status_code} for url `{link}`'
        )


def extract_feed_url_from_html(link: str) -> Optional[str]:
    dest = str(
        CACHE_PATH.joinpath(shasum(link)+'.html')
    )
    try:
        if not isfile(dest):
            download_raw(link, dest)
        sd_html = Html(dest)
        res: str = sd_html.rss_url
        if not res:
            return None
        res = create_full_url(link, res)
        return res
    except Exception:
        print('Error extracting feed from HTML')
    return None


class DownloadFeedResponse:
    def __init__(
            self, feedpath: Optional[Union[Path, Literal['not_cached']]],
            rss_link: Optional[str], failed: bool, error: Optional[str]
    ):
        self.feedpath = feedpath
        self.rss_link = rss_link
        self.failed = failed
        self.error = error


def download_feed(
        link: str, get_cached: bool = False
) -> DownloadFeedResponse:
    confman = ConfManager()

    dest_path: Path = CACHE_PATH.joinpath(shasum(link)+'.rss')
    if get_cached:
        return DownloadFeedResponse(
            dest_path if isfile(dest_path) else 'not_cached',
            link,
            not isfile(dest_path),
            None
        )
    headers = GET_HEADERS.copy()
    if (
            'last-modified' in confman.nconf.feeds[link].keys() and
            isfile(dest_path)
    ):
        headers['If-Modified-Since'] = \
            confman.nconf.feeds[link]['last-modified']
    try:
        res = requests.get(
            link, headers=headers, allow_redirects=True, timeout=TIMEOUT
        )
    except requests.exceptions.ConnectTimeout:
        return DownloadFeedResponse(
            None,
            link,
            True,
            _('`{0}`: connection timed out').format(link)
        )
    except Exception:
        import traceback
        traceback.print_exc()
        return DownloadFeedResponse(
            None,
            link,
            True,
            _('`{0}` might not be a valid address').format(link)
        )
    if 'last-modified' in res.headers.keys():
        # TODO fix when switching to non-json feeds in gsettings
        feeds: dict = confman.nconf.feeds
        feeds[link]['last-modified'] = res.headers['last-modified']
        confman.nconf.feeds = feeds

    def handle_200():
        if (
                'last-modified' not in res.headers.keys() and
                'last-modified' in confman.nconf.feeds[link].keys()
        ):
            feeds: dict = confman.nconf.feeds
            feeds[link].pop('last-modified')
            confman.nconf.feeds = feeds
        with open(dest_path, 'wb') as fd:
            fd.write(res.content)  # res.text is str, res.content is bytes
        to_unicode(dest_path)
        return DownloadFeedResponse(
            dest_path, link, False, None
        )

    def handle_304(): return DownloadFeedResponse(
        dest_path, link, False, None
    )

    def handle_301_302():
        n_link = res.headers.get('location', link)
        feeds: dict = confman.nconf.feeds
        feeds[n_link] = feeds[link]
        feeds.pop(link)
        confman.nconf.feeds = feeds
        return download_feed(n_link)

    def handle_everything_else():
        return DownloadFeedResponse(
            None,
            link,
            True,
            _('Error downloading `{0}`, code `{1}`').format(
                link, res.status_code
            )
        )

    handlers = {
        200: handle_200, 304: handle_304,
        301: handle_301_302, 302: handle_301_302
    }
    return handlers.get(res.status_code, handle_everything_else)()



from typing import Dict, List, Literal, cast
from gfeeds.gsettings_wrapper import GSETTINGS_TYPES, GsettingsWrapper


class ConfMapper:
    def __init__(self, gsw: GsettingsWrapper):
        self.gsw = gsw

    @property
    def feeds(self) -> Dict[
            str,
            Dict[Literal['tags', 'last-modified'], List[str] | str]
    ]:
        return cast(dict, self.gsw.get('feeds'))

    @feeds.setter
    def feeds(self, nval: dict):
        self.set('feeds', nval)

    @property
    def dark_mode(self) -> bool:
        return cast(bool, self.gsw.raw_get('dark-mode'))

    @dark_mode.setter
    def dark_mode(self, nval: bool):
        self.set('dark_mode', nval)

    @property
    def reader_theme(self) -> Literal['auto', 'light', 'dark']:
        return cast(
            Literal['auto', 'light', 'dark'], self.gsw.raw_get('reader-theme')
        )

    @reader_theme.setter
    def reader_theme(self, nval: Literal['auto', 'light', 'dark']):
        self.set('reader_theme', nval)

    @property
    def new_first(self) -> bool:
        return cast(bool, self.gsw.raw_get('new-first'))

    @new_first.setter
    def new_first(self, nval: bool):
        self.set('new_first', nval)

    @property
    def window_height(self) -> int:
        return cast(int, self.gsw.raw_get('window-height'))

    @window_height.setter
    def window_height(self, nval: int):
        self.set('window_height', nval)

    @property
    def window_width(self) -> int:
        return cast(int, self.gsw.raw_get('window-width'))

    @window_width.setter
    def window_width(self, nval: int):
        self.set('window_width', nval)

    @property
    def max_article_age_days(self) -> int:
        return cast(int, self.gsw.raw_get('max-article-age-days'))

    @max_article_age_days.setter
    def max_article_age_days(self, nval: int):
        self.set('max_article_age_days', nval)

    @property
    def enable_js(self) -> bool:
        return cast(bool, self.gsw.raw_get('enable-js'))

    @enable_js.setter
    def enable_js(self, nval: bool):
        self.set('enable_js', nval)

    @property
    def max_refresh_threads(self) -> int:
        return cast(int, self.gsw.raw_get('max-refresh-threads'))

    @max_refresh_threads.setter
    def max_refresh_threads(self, nval: int):
        self.set('max_refresh_threads', nval)

    @property
    def read_items(self) -> List[str]:
        return cast(List[str], self.gsw.raw_get('read-items'))

    @read_items.setter
    def read_items(self, nval: List[str]):
        self.set('read_items', nval)

    @property
    def show_read_items(self) -> bool:
        return cast(bool, self.gsw.raw_get('show-read-items'))

    @show_read_items.setter
    def show_read_items(self, nval: bool):
        self.set('show_read_items', nval)

    @property
    def show_empty_feeds(self) -> bool:
        return cast(bool, self.gsw.raw_get('show-empty-feeds'))

    @show_empty_feeds.setter
    def show_empty_feeds(self, nval: bool):
        self.set('show_empty_feeds', nval)

    @property
    def full_article_title(self) -> bool:
        return cast(bool, self.gsw.raw_get('full-article-title'))

    @full_article_title.setter
    def full_article_title(self, nval: bool):
        self.set('full_article_title', nval)

    @property
    def default_view(self) -> Literal['webview', 'reader', 'feedcont']:
        return cast(
            Literal['webview', 'reader', 'feedcont'],
            self.gsw.raw_get('default-view')
        )

    @default_view.setter
    def default_view(self, nval: Literal['webview', 'reader', 'feedcont']):
        self.set('default_view', nval)

    @property
    def open_links_externally(self) -> bool:
        return cast(bool, self.gsw.raw_get('open-links-externally'))

    @open_links_externally.setter
    def open_links_externally(self, nval: bool):
        self.set('open_links_externally', nval)

    @property
    def full_feed_name(self) -> bool:
        return cast(bool, self.gsw.raw_get('full-feed-name'))

    @full_feed_name.setter
    def full_feed_name(self, nval: bool):
        self.set('full_feed_name', nval)

    @property
    def refresh_on_startup(self) -> bool:
        return cast(bool, self.gsw.raw_get('refresh-on-startup'))

    @refresh_on_startup.setter
    def refresh_on_startup(self, nval: bool):
        self.set('refresh_on_startup', nval)

    @property
    def tags(self) -> List[str]:
        return cast(List[str], self.gsw.raw_get('tags'))

    @tags.setter
    def tags(self, nval: List[str]):
        self.set('tags', nval)

    @property
    def open_youtube_externally(self) -> bool:
        return cast(bool, self.gsw.raw_get('open-youtube-externally'))

    @open_youtube_externally.setter
    def open_youtube_externally(self, nval: bool):
        self.set('open_youtube_externally', nval)

    @property
    def media_player(self) -> str:
        return cast(str, self.gsw.raw_get('media-player'))

    @media_player.setter
    def media_player(self, nval: str):
        self.set('media_player', nval)

    @property
    def show_thumbnails(self) -> bool:
        return cast(bool, self.gsw.raw_get('show-thumbnails'))

    @show_thumbnails.setter
    def show_thumbnails(self, nval: bool):
        self.set('show_thumbnails', nval)

    @property
    def use_experimental_listview(self) -> bool:
        return cast(bool, self.gsw.raw_get('use-experimental-listview'))

    @use_experimental_listview.setter
    def use_experimental_listview(self, nval: bool):
        self.set('use_experimental_listview', nval)

    @property
    def auto_refresh_enabled(self) -> bool:
        return cast(bool, self.gsw.raw_get('auto-refresh-enabled'))

    @auto_refresh_enabled.setter
    def auto_refresh_enabled(self, nval: bool):
        self.set('auto_refresh_enabled', nval)

    @property
    def notify_new_articles(self) -> bool:
        return cast(bool, self.gsw.raw_get('notify-new-articles'))

    @notify_new_articles.setter
    def notify_new_articles(self, nval: bool):
        self.set('notify_new_articles', nval)

    @property
    def auto_refresh_time_seconds(self) -> int:
        return cast(int, self.gsw.raw_get('auto-refresh-time-seconds'))

    @auto_refresh_time_seconds.setter
    def auto_refresh_time_seconds(self, nval: int):
        self.set('auto_refresh_time_seconds', nval)

    @property
    def enable_adblock(self) -> bool:
        return cast(bool, self.gsw.raw_get('enable-adblock'))

    @enable_adblock.setter
    def enable_adblock(self, nval: bool):
        self.set('enable_adblock', nval)

    @property
    def blocklist_last_update(self) -> float:
        return cast(float, self.gsw.raw_get('blocklist-last-update'))

    @blocklist_last_update.setter
    def blocklist_last_update(self, nval: float):
        self.set('blocklist_last_update', nval)

    @property
    def webview_zoom(self) -> float:
        return cast(float, self.gsw.raw_get('webview-zoom'))

    @webview_zoom.setter
    def webview_zoom(self, nval: float):
        self.set('webview_zoom', nval)

    @property
    def font_use_system_for_titles(self) -> bool:
        return cast(bool, self.gsw.raw_get('font-use-system-for-titles'))

    @font_use_system_for_titles.setter
    def font_use_system_for_titles(self, nval: bool):
        self.set('font_use_system_for_titles', nval)

    @property
    def font_use_system_for_paragraphs(self) -> bool:
        return cast(bool, self.gsw.raw_get('font-use-system-for-paragraphs'))

    @font_use_system_for_paragraphs.setter
    def font_use_system_for_paragraphs(self, nval: bool):
        self.set('font_use_system_for_paragraphs', nval)

    @property
    def font_titles_custom(self) -> str:
        return cast(str, self.gsw.raw_get('font-titles-custom'))

    @font_titles_custom.setter
    def font_titles_custom(self, nval: str):
        self.set('font_titles_custom', nval)

    @property
    def font_paragraphs_custom(self) -> str:
        return cast(str, self.gsw.raw_get('font-paragraphs-custom'))

    @font_paragraphs_custom.setter
    def font_paragraphs_custom(self, nval: str):
        self.set('font_paragraphs_custom', nval)

    @property
    def font_monospace_custom(self) -> str:
        return cast(str, self.gsw.raw_get('font-monospace-custom'))

    @font_monospace_custom.setter
    def font_monospace_custom(self, nval: str):
        self.set('font_monospace_custom', nval)

    def set(self, key: str, val: GSETTINGS_TYPES):
        self.gsw.set(key, val)



from gfeeds.util.download_manager import extract_feed_url_from_html
from typing import Optional
from os.path import isfile
from syndom import Feed as SynDomFeed


def get_feed_link_from_file(feed_path: str) -> Optional[str]:
    if not isfile(feed_path):
        return None
    try:
        sd_feed = SynDomFeed(str(feed_path))
    except Exception:
        import traceback
        traceback.print_exc()
        print(f'Error parsing feed `{feed_path}`')
        return None
    res = sd_feed.get_rss_url()
    if res:
        return res
    url = sd_feed.get_url()
    if url:
        return extract_feed_url_from_html(url)
    items = sd_feed.get_items()
    if len(items) <= 0:
        return None
    item_url = items[0].get_url()
    if item_url:
        return extract_feed_url_from_html(item_url)
    return None



from gettext import gettext as _
from typing import TYPE_CHECKING, List, Optional
from bs4 import BeautifulSoup
from gi.repository import GObject, GLib
from dateutil.tz import gettz
from datetime import datetime, timezone
from dateutil.parser import parse as dateparse
from gfeeds.util.get_thumb import get_thumb
from gfeeds.confManager import ConfManager
import pytz
if TYPE_CHECKING:
    from gfeeds.feed import Feed


class FeedItem(GObject.Object):
    __gsignals__ = {
        'changed': (
            GObject.SignalFlags.RUN_FIRST, None, (str,)
        )
    }

    def __init__(self, sd_item, parent_feed: 'Feed'):
        self.confman = ConfManager()
        self.parent_feed = parent_feed
        self.__sd_item = sd_item
        title = self.__sd_item.get_title()
        self.__title = (
            BeautifulSoup(title, features='lxml').text
            if title and '</' in title
            else title
        )
        self.__link = self.__sd_item.get_url()
        self.pub_date_str = self.__sd_item.get_pub_date()
        # fallback to avoid errors
        self.__pub_date = datetime.now(timezone.utc)

        # used to identify article for read/unread and thumbs cache
        self.identifier = self.__link or (self.__title + self.pub_date_str)
        self.__read = self.identifier in self.confman.nconf.read_items

        self.__author_url = self.__sd_item.get_author_url()
        self.__author_name = self.__sd_item.get_author_name()

        self.content = self.__sd_item.get_content()

        try:
            self.__pub_date = dateparse(self.pub_date_str, tzinfos={
                'UT': gettz('GMT'),
                'EST': -18000,
                'EDT': -14400,
                'CST': -21600,
                'CDT': -18000,
                'MST': -25200,
                'MDT': -21600,
                'PST': -28800,
                'PDT': -25200
            })
            if not self.__pub_date.tzinfo:
                self.__pub_date = pytz.UTC.localize(self.__pub_date)
        except Exception:
            print(_(
                'Error: unable to parse datetime {0} for feeditem {1}'
            ).format(self.pub_date_str, self))

        self.__image_url = sd_item.get_img_url()
        # sidebar row will try to async get an image from html if above failed
        super().__init__()

    @GObject.Property(type=str)
    def title(self) -> str:  # type: ignore
        return self.__title

    @GObject.Property()
    def pub_date(self) -> datetime:  # type: ignore
        return self.__pub_date

    @GObject.Property(type=str)
    def link(self) -> str:  # type: ignore
        return self.__link

    @GObject.Property(type=bool, default=False)
    def read(self) -> bool:  # type: ignore
        return self.__read

    @read.setter
    def read(self, n_read: bool):
        self.__set_read(n_read)

    @GObject.Property(type=str)
    def image_url(self) -> str:  # type: ignore
        return self.__image_url

    @GObject.Property(type=str)
    def author_url(self) -> str:  # type: ignore
        return self.__author_url

    @GObject.Property(type=str)
    def author_name(self) -> str:  # type: ignore
        return self.__author_name

    @image_url.setter
    def image_url(self, n_image_url: str):
        self.__image_url = n_image_url

    def set_thumb_from_link(self) -> Optional[str]:
        image_url = get_thumb(self.__link)

        def cb(url):
            self.image_url = url

        GLib.idle_add(cb, image_url)
        return image_url

    def __set_read(self, read):
        if read == self.__read:
            return
        self.parent_feed.unread_count += -1 if read else 1
        read_items: List[str] = self.confman.nconf.read_items  # type: ignore
        if read and self.identifier not in read_items:
            read_items.append(self.identifier)
        elif not read and self.identifier in read_items:
            read_items.remove(self.identifier)
        self.confman.nconf.read_items = read_items
        self.__read = read

    def __repr__(self):
        return 'FeedItem Object `{0}` from Feed {1}'.format(
            self.__title,
            self.parent_feed.title
        )

    def to_dict(self) -> dict:
        return {
            'title': self.title,
            'link': self.link,
            'identifier': self.identifier,
            'pub_date': self.pub_date,
            'image_url': self.image_url,
            'content': self.content,
            'parent_feed': self.parent_feed.to_dict(),
        }



from gi.repository import Gtk, Adw
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.feeds_view import (
    FeedsViewAllListboxRow, FeedsViewListbox, FeedsViewTagListboxRow
)
from gfeeds.tag_store import TagObj


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/filter_view.ui')
class FilterView(Adw.Bin):
    __gtype_name__ = 'FilterView'
    all_listbox = Gtk.Template.Child()
    tags_listbox = Gtk.Template.Child()
    feeds_listbox_bin = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        self.feedman = FeedsManager()
        self.confman = ConfManager()
        self.feeds_listbox = FeedsViewListbox(False)
        self.feeds_listbox.connect(
            'row-activated', self.on_feeds_row_activated
        )
        self.feeds_listbox.connect(
            'row-selected', self.on_feed_row_selected
        )
        self.feeds_listbox_bin.set_child(self.feeds_listbox)
        self.all_listbox_row = FeedsViewAllListboxRow()
        self.all_listbox.append(self.all_listbox_row)
        self.all_listbox.select_row(self.all_listbox_row)
        self.tags_listbox.bind_model(
            self.feedman.tag_store, self.__create_tag_row, None
        )

    def __create_tag_row(self, tag: TagObj, *args) -> FeedsViewTagListboxRow:
        row = FeedsViewTagListboxRow(tag)
        return row

    @Gtk.Template.Callback()
    def on_all_row_activated(self, *_):
        for lb in (self.tags_listbox, self.feeds_listbox):
            lb.select_row(None)
        self.confman.emit('gfeeds_filter_changed', None)

    @Gtk.Template.Callback()
    def on_tags_row_activated(self, _, row):
        for lb in (self.all_listbox, self.feeds_listbox):
            lb.select_row(None)
        self.confman.emit('gfeeds_filter_changed', [row.title])

    def on_feeds_row_activated(self, _, row):
        for lb in (self.all_listbox, self.tags_listbox):
            lb.select_row(None)
        self.confman.emit('gfeeds_filter_changed', row.feed)

    def on_feed_row_selected(self, _, row):
        if not row and not self.tags_listbox.get_selected_row():
            self.all_listbox.select_row(self.all_listbox_row)



from gi.repository import Adw, GLib, Gdk, Gio
from threading import Thread
from pathlib import Path
from os.path import isfile
from PIL import Image
from typing import Union
from gfeeds.util.paths import THUMBS_CACHE_PATH


def make_thumb(path, width: int, height: int = 1000) -> Union[str, None]:
    if not path:
        return None
    if not isinstance(path, Path):
        path = Path(path)
    dest = THUMBS_CACHE_PATH.joinpath(f'{width}x{height}_{path.name}_v2')
    if dest.is_file():
        return str(dest)
    try:
        with Image.open(path) as thumb:
            thumb = Image.open(path)
            thumb.resize((width, height)).save(dest, 'PNG')
            # thumb.thumbnail((width, height), Image.ANTIALIAS)
            # thumb.save(dest, 'PNG')
        return str(dest)
    except IOError:
        print(f'Error creating thumbnail for image `{path}`')
        return None


_textures_cache = dict()


class SimpleAvatar(Adw.Bin):
    def __init__(self):
        super().__init__()
        self.avatar = Adw.Avatar(size=32, show_initials=True)
        self.set_child(self.avatar)

    def set_image(self, title, image=None):
        self.avatar.set_text(title)
        if not image:
            return
        if not isfile(image):
            _textures_cache[image] = None

        def cb(texture, add_to_cache=False):
            if add_to_cache:
                _textures_cache[image] = texture
            self.avatar.set_custom_image(texture)

        if image in _textures_cache.keys():
            cached = _textures_cache[image]
            if not cached:
                return
            GLib.idle_add(cb, cached)
            return

        def af():
            gio_file = Gio.File.new_for_path(image)
            try:
                texture = Gdk.Texture.new_from_file(gio_file)
            except Exception:
                print(
                    'SimpleAvatar: '
                    'Error creating texture for `{0}` (title `{1}`)'.format(
                        image, title
                    )
                )
                texture = None
            GLib.idle_add(cb, texture, True)

        Thread(target=af, daemon=True).start()



from typing import Any, Callable
from gi.repository import Gio

__HOSTNAME = 'gnome.org'


def is_online(callback: Callable[[bool], Any]):

    def __async_cb(src, a_res, data):
        res = False
        try:
            res = Gio.NetworkMonitor.get_default().can_reach_finish(a_res)
        except Exception:
            import traceback
            traceback.print_exc()
            pass
        callback(res)

    Gio.NetworkMonitor.get_default().can_reach_async(
        Gio.NetworkAddress(hostname=__HOSTNAME),
        None,
        __async_cb,
        None
    )



from gfeeds.confManager import ConfManager
from gi.repository import Gtk, Adw, Gio, GObject
from typing import Optional, Union, List, Callable


class MActionRow(Adw.ActionRow):
    def __init__(self, title: str, subtitle: Optional[str] = None, **kwargs):
        self.title = title
        self.subtitle = subtitle
        super().__init__(
            title=self.title,
            title_lines=0, subtitle_lines=0,
            **kwargs
        )
        if self.subtitle:
            self.set_subtitle(self.subtitle)


class PreferencesButtonRow(MActionRow):
    """
    A preferences row with a button
    title: the title shown
    button_label: a label to show inside the button
    onclick: the function that will be called when the button is pressed
    subtitle: an optional subtitle to be shown
    button_style_class: the style class of the button.
        Common options: `suggested-action`, `destructive-action`
    signal: an optional signal to let ConfManager emit when the button is
        pressed
    """
    def __init__(
            self, title: str, button_label: str,
            onclick: Callable, subtitle: Optional[str] = None,
            button_style_class: Optional[str] = None,
            signal: Optional[str] = None
    ):
        super().__init__(title, subtitle)
        self.button_label = button_label
        self.confman = ConfManager()
        self.signal = signal
        self.onclick = onclick

        self.button = Gtk.Button(
            label=self.button_label, valign=Gtk.Align.CENTER
        )
        if button_style_class:
            self.button.get_style_context().add_class(button_style_class)
        self.button.connect('clicked', self.on_button_clicked)
        self.add_suffix(self.button)

    def on_button_clicked(self, btn):
        self.onclick(self.confman)
        if self.signal:
            self.confman.emit(self.signal)


class PreferencesEntryRow(MActionRow):
    """
    A preferences row with an entry
    title: the title shown
    conf_key: the key of the configuration dictionary/json in ConfManager
    subtitle: an optional subtitle to be shown
    onchange: an optional function that will be called when the entry changes
    signal: an optional signal to let ConfManager emit when the entry changes
    """
    def __init__(
            self, title: str, conf_key: str, subtitle: Optional[str] = None,
            onchange: Optional[Callable] = None, signal: Optional[str] = None
    ):
        super().__init__(title, subtitle)
        self.conf_key = conf_key
        self.confman = ConfManager()
        self.signal = signal
        self.onchange = onchange

        self.entry = Gtk.Entry(valign=Gtk.Align.CENTER)
        self.entry.set_text(self.confman.conf[self.conf_key])
        self.entry.connect('changed', self.on_entry_changed)
        self.add_suffix(self.entry)

    def on_entry_changed(self, *_):
        self.confman.conf[self.conf_key] = self.entry.get_text().strip()
        if self.onchange is not None:
            self.onchange(self.confman)
        if self.signal:
            self.confman.emit(self.signal)


class PreferencesSpinButtonRow(MActionRow):
    """
    A preferences row with a spin button
    title: the title shown
    min_v: minimum num value
    max_v: maximum num value
    conf_key: the key of the configuration dictionary/json in ConfManager
    subtitle: an optional subtitle to be shown
    signal: an optional signal to let ConfManager emit when the value changes
    """

    def __init__(
            self, title: str, min_v: int, max_v: int, conf_key: str,
            subtitle: Optional[str] = None, signal: Optional[str] = None
    ):
        super().__init__(title, subtitle)
        self.confman = ConfManager()
        self.signal = signal
        self.conf_key = conf_key

        self.adjustment = Gtk.Adjustment.new(
            self.confman.conf[self.conf_key],  # initial value
            min_v,  # minimum value
            max_v,  # maximum value
            1,  # step increment
            10,  # page increment (page up, page down? large steps anyway)
            0
        )

        self.spin_button = Gtk.SpinButton(
            adjustment=self.adjustment, valign=Gtk.Align.CENTER
        )
        self.spin_button.connect('value-changed', self.on_value_changed)
        self.add_suffix(self.spin_button)

    def on_value_changed(self, *_):
        self.confman.conf[self.conf_key] = self.spin_button.get_value_as_int()
        if self.signal:
            self.confman.emit(self.signal, self.confman.conf[self.conf_key])


class PreferencesComboRow(Adw.ComboRow):
    """
    A preferences row with a combo box
    title: the title shown
    values: a list of acceptable values
    value_names: a list of user facing names for the values provided above
    conf_key: the key of the configuration dictionary/json in ConfManager
    subtitle: an optional subtitle to be shown
    signal: an optional signal to let ConfManager emit when the value changes
    """

    class ItemWrapper(GObject.Object):
        def __init__(self, name: str, value: str):
            super().__init__()
            self.name = name
            self.value = value

    def __init__(
            self, title: str, values: List[str], value_names: List[str],
            conf_key: str, subtitle: Optional[str] = None,
            signal: Optional[str] = None
    ):
        self.confman = ConfManager()
        self.signal = signal
        self.conf_key = conf_key
        self.list_store = Gio.ListStore(
            item_type=PreferencesComboRow.ItemWrapper
        )
        self.items_l = list()
        for name, value in zip(value_names, values):
            i = PreferencesComboRow.ItemWrapper(name, value)
            self.items_l.append(i)
            self.list_store.append(i)
        self.factory = Gtk.SignalListItemFactory()
        self.factory.connect('setup', self._on_setup_listitem)
        self.factory.connect('bind', self._on_bind_listitem)

        self.title = title
        self.subtitle = subtitle

        super().__init__(
            model=self.list_store, factory=self.factory, title=title
        )
        if self.subtitle:
            self.set_subtitle(self.subtitle)

        self.set_selected(values.index(self.confman.conf[self.conf_key]))

        self.connect('notify::selected-item', self.on_selection_changed)

    def _on_setup_listitem(
            self, factory: Gtk.ListItemFactory, list_item: Gtk.ListItem
    ):
        label = Gtk.Label()
        list_item.set_child(label)
        list_item.row_w = label

    def _on_bind_listitem(
            self, factory: Gtk.ListItemFactory, list_item: Gtk.ListItem
    ):
        label = list_item.get_child()
        label.set_text(list_item.get_item().name)

    def on_selection_changed(self, *args):
        value = self.get_selected_item().value
        if value is not None:
            self.confman.conf[self.conf_key] = value
            if self.signal:
                self.confman.emit(self.signal)


class PreferencesToggleRow(MActionRow):
    """
    A preferences row with a toggle
    title: the title shown
    conf_key: the key of the configuration dictionary/json in ConfManager
    subtitle: an optional subtitle to be shown
    signal: an optional signal to let ConfManager emit when the configuration
        is set
    """
    def __init__(
            self, title: str, conf_key: str, subtitle: Optional[str] = None,
            signal: Optional[str] = None
    ):
        super().__init__(title, subtitle)
        self.confman = ConfManager()
        self.conf_key = conf_key
        self.signal = signal

        self.toggle = Gtk.Switch(valign=Gtk.Align.CENTER)
        self.toggle.set_active(self.confman.conf[self.conf_key])
        self.toggle.connect('state-set', self.on_toggle_state_set)
        self.add_suffix(self.toggle)
        self.set_activatable_widget(self.toggle)

    def on_toggle_state_set(self, toggle, state):
        self.confman.conf[self.conf_key] = state
        if self.signal is not None:
            self.confman.emit(self.signal)


class PreferencesFontChooserRow(MActionRow):
    """
    A preference row with a font chooser button
    """
    def __init__(
        self, title: str, conf_key: str, subtitle: Optional[str] = None,
        signal: Optional[str] = None
    ):
        super().__init__(title, subtitle)
        self.confman = ConfManager()
        self.conf_key = conf_key
        self.signal = signal

        self.font_btn = Gtk.FontButton(
            title=self.title, modal=True, use_size=False,
            use_font=True, font=self.confman.conf[self.conf_key],
            valign=Gtk.Align.CENTER, level=Gtk.FontChooserLevel.FAMILY
        )
        self.font_btn.connect('font-set', self.on_font_set)
        self.add_suffix(self.font_btn)
        self.set_activatable_widget(self.font_btn)

    def on_font_set(self, *args):
        n_font = ' '.join(self.font_btn.get_font().split(' ')[:-1]).strip()
        self.confman.conf[self.conf_key] = n_font
        if self.signal:
            self.confman.emit(self.signal)


class MPreferencesGroup(Adw.PreferencesGroup):
    def __init__(
            self, title: str,
            rows: List[Union[MActionRow, PreferencesComboRow, Adw.ActionRow]]
    ):
        self.title = title
        self.rows = rows
        super().__init__(title=self.title)
        for row in self.rows:
            self.add(row)


class MPreferencesPage(Adw.PreferencesPage):
    def __init__(
            self, title: str, pref_groups: List[MPreferencesGroup],
            icon_name: Optional[str] = None
    ):
        self.title = title
        self.icon_name = icon_name
        self.pref_groups = pref_groups
        super().__init__(title=self.title)
        if self.icon_name:
            self.set_icon_name(self.icon_name)
        for group in self.pref_groups:
            self.add(group)



from gi.repository import Gtk, Gdk, Gio, GLib
from gfeeds.confManager import ConfManager
from threading import Thread
from math import ceil


class PictureView(Gtk.Widget):
    def __init__(self, path):
        super().__init__(
            overflow=Gtk.Overflow.HIDDEN, vexpand=True,
            hexpand=True,
            # valign=Gtk.Align.CENTER
        )
        self.get_style_context().add_class('card')
        self.confman = ConfManager()
        self.texture = None
        self.set_file(path)

    def set_file(self, path):
        self.path = path
        if self.path is None:
            self.texture = None
            return
        gio_file = Gio.File.new_for_path(self.path)

        def af():
            try:
                self.texture = Gdk.Texture.new_from_file(gio_file)
            except Exception:
                print(
                    f'PictureView: Error creating texture for `{self.path}`'
                )
                self.texture = None
            GLib.idle_add(cb)

        def cb():
            if self.texture is None:
                return
            # 270 is the pic natural width with leaflet unfolded
            cw, ch = self.texture.compute_concrete_size(
                270, 0, 1200, 1200
            )
            self.set_size_request(-1, ceil(ch))
            self.queue_draw()
            self.queue_resize()

        Thread(target=af, daemon=True).start()

    def do_get_request_mode(self, *args):
        return Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH

    def do_snapshot(self, snapshot):
        if self.texture is None:
            return
        width = self.get_width()
        height = width / self.texture.get_intrinsic_aspect_ratio()
        self.texture.snapshot(
            snapshot,
            width, height
        )

    def do_measure(self, orientation, for_size):
        if orientation == Gtk.Orientation.VERTICAL:  # get height
            if for_size == -1:
                return (0, 0, -1, -1)
            if not self.texture:
                return (0, 0, -1, -1)
            cw, ch = self.texture.compute_concrete_size(
                for_size, 0, 1200, 1200
            )
            ch = max(min(ceil(ch), 1200), 1)
            return (0, ch, -1, -1)
        else:  # get width
            if for_size == -1:
                return (0, 1200, -1, -1)
            return (0, 1200, -1, -1)



from gi.repository import GObject, Gtk, Gio
from typing import Optional
from gfeeds.confManager import ConfManager
from gfeeds.feed import Feed


class FeedStore(Gtk.FilterListModel):
    __gsignals__ = {
        'item-removed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT,)
        )
    }

    def __init__(self):
        self.sorter = Gtk.CustomSorter()
        self.sorter.set_sort_func(self._sort_func)
        self.filter = Gtk.CustomFilter()
        self.filter.set_filter_func(self._filter_func)
        self.list_store = Gio.ListStore(item_type=Feed)
        self.sort_store = Gtk.SortListModel(
            model=self.list_store, sorter=self.sorter
        )
        self.confman = ConfManager()
        self.confman.connect(
                'gfeeds_show_empty_feeds_changed',
                lambda *_: self.invalidate_filter()
        )
        # Hiding read articles can result in empty feeds which should
        # be hidden
        self.confman.connect(
                'gfeeds_show_read_changed',
                lambda *_: self.invalidate_filter()
        )
        super().__init__(model=self.sort_store, filter=self.filter)

    def _sort_func(self, feed1: Feed, feed2: Feed, *_) -> int:
        return -1 if feed1.title.lower() < feed2.title.lower() else 1

    def invalidate_sort(self):
        self.sorter.set_sort_func(self._sort_func)

    def _filter_func(self, feed: Feed, *_) -> bool:
        res = True
        if not self.confman.nconf.show_empty_feeds:
            if self.confman.nconf.show_read_items:
                res = res and len(feed.items) > 0
            else:
                res = res and feed.unread_count > 0
        return res

    def invalidate_filter(self):
        self.filter.set_filter_func(self._filter_func)

    def empty(self):
        return self.list_store.remove_all()

    def add_feed(self, n_feed: Feed):
        self.list_store.append(n_feed)
        n_feed.connect(
            'empty_changed', lambda *_: self.invalidate_filter()
        )

    def remove_by_index(self, index: int):
        self.emit('item-removed', self.list_store[index])
        self.list_store.remove(index)

    def remove_feed(self, to_rm: Feed):
        for i, feed in enumerate(self.list_store):
            if feed.title == to_rm.title and feed.rss_link == to_rm.rss_link:
                self.remove_by_index(i)
                return

    def get_feed(self, identifier: str) -> Optional[Feed]:
        for f in self.list_store:
            if f.rss_link + f.title == identifier:
                return f
        return None



from gi.repository import Gtk
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from gfeeds.articles_listview import ArticlesListView, ArticlesListBox


class LoadingRevealer(Gtk.Revealer):
    def __init__(self):
        super().__init__(
            transition_type=Gtk.RevealerTransitionType.CROSSFADE,
            reveal_child=False, vexpand=False, hexpand=True,
            valign=Gtk.Align.START, halign=Gtk.Align.FILL
        )
        self.progress_bar = Gtk.ProgressBar(hexpand=True, vexpand=False)
        self.progress_bar.get_style_context().add_class('osd')
        self.set_child(self.progress_bar)

    def set_running(self, state: bool):
        self.set_reveal_child(state)

    def set_progress(self, progress: float):
        '''progress: float between 0.0 and 1.0'''

        self.progress_bar.set_fraction(progress)


class GFeedsSidebar(Gtk.Overlay):
    __gtype_name__ = 'GFeedsSidebar'

    def __init__(self):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = FeedsManager()

        self.listview_sw = (
            ArticlesListView()
            if self.confman.nconf.use_experimental_listview
            else ArticlesListBox()
        )
        self.empty = self.listview_sw.empty
        self.populate = self.listview_sw.populate
        self.loading_revealer = LoadingRevealer()

        self.set_child(self.listview_sw)
        self.add_overlay(self.loading_revealer)

        self.feedman.connect(
            'feedmanager_refresh_start',
            self.on_refresh_start
        )
        self.feedman.connect(
            'feedmanager_refresh_end',
            self.on_refresh_end
        )
        self.feedman.connect(
            'feedmanager_feeds_loaded_changed',
            self.on_feeds_loaded_changed
        )

    def on_feeds_loaded_changed(self, _, progress: float):
        self.loading_revealer.set_progress(progress)

    def on_refresh_start(self, *_):
        self.loading_revealer.set_running(True)

    def on_refresh_end(self, *_):
        self.listview_sw.all_items_changed()
        self.loading_revealer.set_running(False)

    def set_search(self, search_term):
        self.listview_sw.set_search_term(search_term)

    def select_next_article(self, *_):
        self.listview_sw.select_next()

    def select_prev_article(self, *_):
        self.listview_sw.select_prev()



from gettext import gettext
from typing import List, Tuple, Union
from gfeeds.feed_item import FeedItem
from gfeeds.util.readability_wrapper import RDoc
import pygments
import pygments.lexers
from lxml.html import (
    fromstring as html_fromstring,
    tostring as html_tostring,
    HtmlElement
)
from pygments.formatters import HtmlFormatter
from gfeeds.util.reader_mode_style import get_css


# Thanks to Eloi Rivard (azmeuk) for the contribution on the media block
def _build_media_text(title: str, content: str) -> str:
    return '''
        <p>
            <strong>{0}:</strong>
            {1}
        </p>'''.format(title, content.replace("\n", "<br />"))


def _build_media_link(title: str, content: str, link: str):
    return _build_media_text(
        title, f'<a href="{link}">{content}</a>'
    )


# thumbnails aren't supposed to be links, images can be?
# funnily enough, using `#` as a link opens the main content url
# that's because the webkitview sets the base url to the feed item link
def _build_media_img(title, imgurl, link='#') -> str:
    return _build_media_link(
        title, f'<br /><img src="{imgurl}" />', link
    )


def build_syntax_highlight(root: HtmlElement) -> Tuple[str, HtmlElement]:
    syntax_highlight_css = ''
    code_nodes: List[HtmlElement] = root.xpath(
        '//pre/code'
    )
    lexer = None
    for code_node in code_nodes:
        classes = code_node.attrib.get('class', '').split(' ')
        for klass in classes:
            try:
                lexer = pygments.lexers.get_lexer_by_name(
                    klass.replace('language-', ''),
                    stripall=True,
                )
                break
            except pygments.util.ClassNotFound:
                pass

        code_text = code_node.text_content()
        if not code_text:
            continue

        if lexer is None:
            try:
                lexer = pygments.lexers.guess_lexer(code_text)
            except pygments.util.ClassNotFound:
                continue

        formatter = HtmlFormatter(style='solarized-dark', linenos=False)

        if not syntax_highlight_css:
            syntax_highlight_css = formatter.get_style_defs()

        newtext = pygments.highlight(code_text, lexer, formatter)
        newhtml = html_fromstring(newtext)

        pre_node = code_node.getparent()
        pre_node.getparent().replace(pre_node, newhtml)

    return syntax_highlight_css, root


def build_syntax_highlight_from_raw_html(
        raw_html: Union[str, bytes]
) -> Tuple[str, HtmlElement]:
    return build_syntax_highlight(
        html_fromstring(
            raw_html if isinstance(raw_html, str) else raw_html.decode()
        )
    )


def build_reader_html(
        og_html, feed_item: FeedItem, dark_mode: bool = False
) -> str:
    doc = RDoc(og_html)
    content = doc.summary(html_partial=True)
    syntax_highlight_css, root = build_syntax_highlight_from_raw_html(content)
    content = html_tostring(
        root, encoding='utf-8'
    )
    author_html = ''
    if feed_item.author_name:
        if feed_item.author_url:
            author_html = gettext('Author: <a href="{0}">{1}</a>').format(
                feed_item.author_url, feed_item.author_name
            )
        else:
            author_html = gettext('Author: {0}'.format(feed_item.author_name))
    if not isinstance(content, str):
        content = content.decode()
    return f'''<html>
        <head>
            <meta charset="UTF-8" />
            <style>
                {get_css()}
                {syntax_highlight_css}
            </style>
            <title>{doc.short_title() or feed_item.title}</title>
        </head>
        <body {'class="dark"' if dark_mode else ''} dir=auto>
            <article>
                <h1>{doc.short_title() or feed_item.title}</h1>
                <p>{author_html}</p>
                {
                    f'<img src="{feed_item.image_url}" /> <hr />'
                    if feed_item.image_url
                    and feed_item.image_url not in content
                    else ''
                }
                {content}
            </article>
        </body>
    </html>'''



from gettext import gettext as _
from threading import Thread
from gi.repository import Gtk, GLib, WebKit, GObject, Gio, Adw
from gfeeds.util.build_reader_html import build_reader_html
from gfeeds.confManager import ConfManager
from gfeeds.util.download_manager import DownloadError, download_text
from functools import reduce
from operator import or_
from subprocess import Popen
from datetime import datetime
from typing import Optional, cast
from gfeeds.feed_item import FeedItem
from gfeeds.util.paths import CACHE_PATH, IS_FLATPAK


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/webview.ui')
class GFeedsWebView(Gtk.Stack):
    __gtype_name__ = 'GFeedsWebView'
    webkitview = cast(WebKit.WebView, Gtk.Template.Child())
    loading_bar_revealer = cast(Gtk.Revealer, Gtk.Template.Child())
    loading_bar = cast(Gtk.ProgressBar, Gtk.Template.Child())
    main_view = cast(Gtk.Overlay, Gtk.Template.Child())
    toast_overlay = cast(Adw.ToastOverlay, Gtk.Template.Child())
    link_preview_revealer = cast(Gtk.Revealer, Gtk.Template.Child())
    link_preview_label = cast(Gtk.Label, Gtk.Template.Child())

    __gsignals__ = {
        'gfeeds_webview_load_start': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (str,)
        ),
        'zoom_changed': (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (float,)
        )
    }

    def __init__(self):
        super().__init__()
        self.confman = ConfManager()

        self.webkitview_settings = WebKit.Settings()
        self.apply_webview_settings()

        self.confman.connect(
            'gfeeds_webview_settings_changed',
            self.apply_webview_settings
        )

        self.confman.connect(
            'on_apply_adblock_changed',
            lambda *args: self.on_apply_adblock_changed(False)
        )
        self.confman.connect(
            'on_refresh_blocklist',
            lambda *args: self.on_apply_adblock_changed(True)
        )

        self.content_manager = self.webkitview.get_user_content_manager()
        self.user_content_filter_store = WebKit.UserContentFilterStore.new(
            str(CACHE_PATH.joinpath(
                'webkit_user_content_filter_store'
            ))
        )
        if self.confman.nconf.enable_adblock:
            self.apply_adblock()

        self.webkitview.set_zoom_level(self.confman.nconf.webview_zoom)

        self.new_page_loaded = False
        self.uri = ''
        self.feeditem = None
        self.html = None

    @Gtk.Template.Callback()
    def on_mouse_target_changed(self, webkitview, hit_test_result, modifiers):
        if hit_test_result:
            if hit_test_result.context_is_link():
                self.link_preview_revealer.set_visible(True)
                self.link_preview_revealer.set_reveal_child(True)
                self.link_preview_label.set_text(
                    hit_test_result.get_link_uri()
                )
                return
        self.link_preview_revealer.set_visible(False)
        self.link_preview_revealer.set_reveal_child(False)

    def action_open_media_player(self):
        self.open_url_in_media_player(
            self.feeditem.link if self.feeditem else None
        )

    def on_apply_adblock_changed(self, refresh: bool):
        self.apply_adblock(
            refresh=refresh, remove=not self.confman.nconf.enable_adblock
        )

    def apply_adblock(self, refresh: bool = False, remove: bool = False):
        refresh = refresh or (
            datetime.fromtimestamp(
                self.confman.nconf.blocklist_last_update
            ) - datetime.now()
        ).days >= 10

        if refresh or remove:
            self.content_manager.remove_filter_by_id('blocklist')
        if remove:
            return

        def apply_filter(filter: WebKit.UserContentFilter):
            self.content_manager.add_filter(filter)

        def save_blocklist_cb(caller, res, *args):
            try:
                filter = self.user_content_filter_store.save_finish(res)
                apply_filter(filter)
            except GLib.Error:
                print('Error saving blocklist')

        def download_blocklist_cb(blocklist: str):
            self.user_content_filter_store.save(
                'blocklist', GLib.Bytes.new(blocklist.encode()), None,
                save_blocklist_cb
            )

        def download_blocklist():
            res = download_text(
                'https://easylist-downloads.adblockplus.org/'
                'easylist_min_content_blocker.json'
            )
            now = datetime.now()
            print(f'Downloaded updated blocklist at {now}')
            self.confman.nconf.blocklist_last_update = now.timestamp()
            GLib.idle_add(download_blocklist_cb, res)

        def filter_load_cb(caller, res, *args):
            try:
                filter = self.user_content_filter_store.load_finish(res)
                apply_filter(filter)
                print('Loaded stored blocklist')
            except GLib.Error:
                print('blocklist store not found, downloading...')
                Thread(target=download_blocklist, daemon=True).start()

        if refresh:
            Thread(target=download_blocklist, daemon=True).start()
        else:
            self.user_content_filter_store.load(
                'blocklist', None, filter_load_cb, None
            )

    def change_view_mode(self, target):
        if target == 'webview':
            # if uri is empty force rss content
            if not self.uri:
                if not self.feeditem:
                    return
                return self.load_feeditem(self.feeditem, force_feedcont=True)
            self.webkitview.load_uri(self.uri)
        elif target == 'reader':
            Thread(
                target=self._load_reader_async,
                args=(self.load_reader,), daemon=True
            ).start()
        elif target == 'feedcont':
            self.set_enable_rss_content(True)

    def apply_webview_settings(self, *args):
        self.webkitview_settings.set_enable_javascript(
            self.confman.nconf.enable_js
        )
        self.webkitview_settings.set_enable_smooth_scrolling(True)
        self.webkitview_settings.set_enable_page_cache(True)
        self.webkitview.set_settings(self.webkitview_settings)

    def key_zoom_in(self, *args):
        self.webkitview.set_zoom_level(
            self.webkitview.get_zoom_level()+0.10
        )
        self.on_zoom_changed()

    def key_zoom_out(self, *args):
        self.webkitview.set_zoom_level(
            self.webkitview.get_zoom_level()-0.10
        )
        self.on_zoom_changed()

    def key_zoom_reset(self, *args):
        self.webkitview.set_zoom_level(1.0)
        self.on_zoom_changed()

    def on_zoom_changed(self):
        zoom = self.webkitview.get_zoom_level()
        self.emit('zoom_changed', zoom)
        self.confman.nconf.webview_zoom = zoom

    def show_notif(self, *args):
        toast = Adw.Toast(title=_('Link copied to clipboard!'))
        self.toast_overlay.add_toast(toast)

    def set_enable_rss_content(self, state=True, feeditem=None):
        if feeditem:
            self.feeditem = feeditem
        if state:
            self.load_rss_content(self.feeditem)
        else:
            self.new_page_loaded = True
            self.load_feeditem(
                self.feeditem,
                False
            )

    def load_rss_content(self, feeditem):
        self.set_visible_child(self.main_view)
        self.feeditem = feeditem
        self.uri = feeditem.link
        content = feeditem.content
        if not content:
            content = '<h1><i>'+_(
                'Feed content not available for this article'
                )+'</i></h1>'
        self.html = '<!-- GFEEDS RSS CONTENT --><article>{0}</article>'.format(
            content if '</' in content else content.replace('\n', '<br>')
        )
        self.load_reader()

    def _load_reader_async(self, callback=None, *args):
        # if uri is empty force rss content
        if not self.uri:
            if not self.feeditem:
                return
            return self.load_feeditem(self.feeditem, force_feedcont=True)
        try:
            self.html = download_text(self.uri)
        except DownloadError as err:
            self.html = (
                f'<h1>{_("Error downloading content.")}</h1>'
                f'<h3>{_("Error code:")} {err.download_error_code}</h3>'
            )
        if callback:
            GLib.idle_add(callback)

    def load_feeditem(
            self, feeditem: FeedItem,
            trigger_on_load_start: Optional[bool] = True,
            force_feedcont: Optional[bool] = False
    ):
        self.webkitview.stop_loading()
        uri = feeditem.link
        self.feeditem = feeditem
        self.uri = uri
        self.set_visible_child(self.main_view)
        target = self.confman.nconf.default_view
        # if uri is empty, fallback to rss content
        if not uri or force_feedcont:
            target = 'feedcont'
        if target == 'reader':
            Thread(
                target=self._load_reader_async,
                args=(self.load_reader,),
                daemon=True
            ).start()
            if trigger_on_load_start:
                self.on_load_start()
        elif target == 'feedcont':
            self.on_load_start()
            self.set_enable_rss_content(True, feeditem)
        else:
            self.webkitview.load_uri(uri)
            if trigger_on_load_start:
                self.on_load_start()

    def open_externally(self, *args):
        if self.uri:
            Gio.AppInfo.launch_default_for_uri(
                self.uri
            )

    def on_load_start(self, *args):
        self.new_page_loaded = True
        self.emit('gfeeds_webview_load_start', '')

    @Gtk.Template.Callback()
    def on_load_changed(self, webview, event):
        if event != WebKit.LoadEvent.FINISHED:
            self.loading_bar_revealer.set_reveal_child(True)
        if event == WebKit.LoadEvent.STARTED:
            self.loading_bar.set_fraction(.25)
        elif event == WebKit.LoadEvent.REDIRECTED:
            self.loading_bar.set_fraction(.50)
        elif event == WebKit.LoadEvent.COMMITTED:
            self.loading_bar.set_fraction(.75)
        elif event == WebKit.LoadEvent.FINISHED:
            self.loading_bar.set_fraction(1.0)
            # waits 1 seconds async then hides the loading bar
            GLib.timeout_add_seconds(
                1,
                self.loading_bar_revealer.set_reveal_child, False
            )
            self.new_page_loaded = False
            resource = webview.get_main_resource()
            if resource:
                resource.get_data(None, self._get_data_cb, None)

    def load_reader(self):
        if not self.feeditem:
            return
        dark = False
        if self.confman.nconf.reader_theme == 'auto':
            dark = Adw.StyleManager.get_default().get_dark()
        else:
            dark = self.confman.nconf.reader_theme == 'dark'
        self.webkitview.load_html(build_reader_html(
            self.html,
            self.feeditem,
            dark,
        ), self.uri)

    def _get_data_cb(self, resource, result, __=None):
        self.html = resource.get_data_finish(result)

    def open_url_in_media_player(self, url: Optional[str]):
        if not url:
            return
        cmd_parts = [
            self.confman.nconf.media_player, f"'{url}'"
        ]
        if IS_FLATPAK:
            cmd_parts.insert(0, 'flatpak-spawn --host')
        cmd = ' '.join(cmd_parts)
        Popen(cmd, shell=True)

    @Gtk.Template.Callback()
    def on_decide_policy(self, webView, decision, decisionType):
        if (
                decisionType in
                (
                    WebKit.PolicyDecisionType.NAVIGATION_ACTION,
                    WebKit.PolicyDecisionType.NEW_WINDOW_ACTION
                ) and
                decision.get_navigation_action().get_mouse_button() != 0
        ):
            uri = decision.get_navigation_action().get_request().get_uri()
            if (
                    self.confman.nconf.open_youtube_externally and
                    reduce(or_, [
                        f'://{pfx}' in uri
                        for pfx in [
                            p + 'youtube.com'
                            for p in ('', 'www.', 'm.')
                        ]
                    ])
            ):
                decision.ignore()
                self.open_url_in_media_player(uri)
            else:
                if not self.confman.nconf.open_links_externally:
                    return False
                decision.ignore()
                Gio.AppInfo.launch_default_for_uri(uri)
            return True
        return False



from os import environ
from pathlib import Path


IS_FLATPAK = Path('/.flatpak-info').is_file()
CONF_DIR = Path(
    environ.get('XDG_CONFIG_HOME') or f'{environ.get("HOME")}/.config'
)
CACHE_HOME = Path(
    environ.get('XDG_CACHE_HOME') or f'{environ.get("HOME")}/.cache'
)
CACHE_PATH = CACHE_HOME.joinpath('org.gabmus.gfeeds')
THUMBS_CACHE_PATH = CACHE_PATH.joinpath('thumbnails')
ARTICLE_THUMB_CACHE_PATH = THUMBS_CACHE_PATH.joinpath(
    'article_thumb_cache.json'
)

for p in [
        THUMBS_CACHE_PATH,
        CONF_DIR
]:
    if not p.is_dir():
        p.mkdir(parents=True)



from syndom import Html
from gfeeds.util.create_full_url import create_full_url
from gfeeds.util.download_manager import download_raw
from gfeeds.util.paths import CACHE_PATH
from gfeeds.util.sha import shasum
from os.path import isfile


def get_thumb(link):
    dest = str(
        CACHE_PATH.joinpath(shasum(link)+'.html')
    )
    try:
        if not isfile(dest):
            download_raw(link, dest)
        sd_html = Html(dest)
    except Exception:
        print('Error parsing HTML')
        return None
    res = sd_html.img_url
    if not res:
        return None
    return create_full_url(link, res)



from typing import Callable, List
from gi.repository import Gtk


class Accelerator:
    def __init__(self, combo: str, cb: Callable):
        self.combo = combo
        self.cb = cb


def add_accelerators(
        window: Gtk.Window,
        shortcuts_l: List[Accelerator]
) -> Gtk.ShortcutController:
    shortcut_controller = Gtk.ShortcutController()
    shortcut_controller.set_scope(Gtk.ShortcutScope.GLOBAL)
    for s in shortcuts_l:
        __add_accelerator(
            shortcut_controller, s.combo, s.cb
        )
    window.add_controller(shortcut_controller)
    return shortcut_controller


def __add_accelerator(
        controller: Gtk.ShortcutController, shortcut: str, callback: Callable
):
    if shortcut:
        # res is bool, don't know what it is
        _, key, mod = Gtk.accelerator_parse(shortcut)
        gshcut = Gtk.Shortcut(
            trigger=Gtk.KeyvalTrigger(keyval=key, modifiers=mod),
            action=Gtk.CallbackAction.new(callback)
        )
        controller.add_shortcut(gshcut)


def add_mouse_button_accel(
        widget: Gtk.Widget, function: Callable,
        propagation: Gtk.PropagationPhase = Gtk.PropagationPhase.BUBBLE
) -> Gtk.GestureClick:
    '''Adds an accelerator for mouse btn press for widget to function.
    NOTE: this returns the Gtk.Gesture, you need to keep this around or it
    won't work. Assign it to some random variable and don't let it go out of
    scope'''

    gesture = Gtk.GestureClick.new()
    gesture.set_button(0)
    gesture.set_propagation_phase(propagation)
    gesture.connect('pressed', function)
    widget.add_controller(gesture)
    return gesture


def add_longpress_accel(
        widget: Gtk.Widget, function: Callable,
        propagation: Gtk.PropagationPhase = Gtk.PropagationPhase.BUBBLE
) -> Gtk.GestureLongPress:
    '''Adds an accelerator for mouse btn press for widget to function.
    NOTE: this returns the Gtk.Gesture, you need to keep this around or it
    won't work. Assign it to some random variable and don't let it go out of
    scope'''

    gesture = Gtk.GestureLongPress.new()
    gesture.set_propagation_phase(propagation)
    gesture.set_touch_only(False)
    gesture.connect('pressed', function)
    widget.add_controller(gesture)
    return gesture



from pathlib import Path
from typing import List, Optional
from gettext import gettext as _
from gfeeds.util.get_favicon import get_favicon
from os.path import isfile
from gfeeds.util.paths import THUMBS_CACHE_PATH
from gfeeds.util.sha import shasum
# from bs4 import UnicodeDammit  # TODO: reimplement it!
from syndom import Feed as SynDomFeed, FeedItem as SynDomFeedItem


class FeedParserRes:
    def __init__(
        self,
        is_null: bool = False,
        error: Optional[str] = None,
        sd_feed: SynDomFeed = None,
        rss_link: Optional[str] = None,
        title: Optional[str] = None,
        link: Optional[str] = None,
        description: Optional[str] = None,
        image_url: Optional[str] = None,
        favicon_path: Optional[str] = None,
        raw_entries: List[SynDomFeedItem] = []
    ):
        self.is_null = is_null
        self.error = error
        self.sd_feed = sd_feed
        self.rss_link = rss_link or ''
        self.title = title or ''
        self.link = link or ''
        self.description = description or ''
        self.image_url = image_url
        self.favicon_path = favicon_path
        self.raw_entries = raw_entries

    @property
    def feed_identifier(self) -> str:
        assert self.rss_link
        assert self.title
        return self.rss_link + self.title


def parse_feed(
        feedpath: Optional[Path],
        rss_link_: Optional[str] = None,
        failed: bool = False,
        error: Optional[str] = None
) -> FeedParserRes:
    if failed:
        print(error)
        return FeedParserRes(is_null=True, error=(error or '<NULL ERROR>'))
    sd_feed = None
    try:
        sd_feed = SynDomFeed(str(feedpath))
    except Exception:
        print('Error parsing feed (caught); will try extracting from HTML')
    if sd_feed is None:
        return FeedParserRes(
            is_null=True, error=_(
                'Errors while parsing feed `{0}`, URL: `{1}`'
            ).format(feedpath, rss_link_)
        )
    try:
        title = sd_feed.get_title()
        raw_entries = sd_feed.get_items()
        link = sd_feed.get_url()
        rss_link = rss_link_ or sd_feed.get_rss_url()
        image_url = sd_feed.get_img_url()
        description = sd_feed.get_description()
    except UnicodeDecodeError:
        return FeedParserRes(
            is_null=True,
            error=_(
                'Error decoding unicode data from feed `{0}`, URL: `{1}`'
            ).format(feedpath, rss_link_)
        )
    except Exception:
        return FeedParserRes(
            is_null=True,
            error=_(
                'Error extracting data from feed `{0}`, URL: `{1}`'
            ).format(feedpath, rss_link_)
        )
    if not title and len(raw_entries) == 0:
        # if these conditions are met, there's reason to believe
        # this is not an rss/atom feed
        return FeedParserRes(
            is_null=False, error=_(
                '`{0}` may not be an RSS or Atom feed'
            ).format(rss_link_)
        )
    if not title:
        title = rss_link
    favicon_path: Optional[str] = str(THUMBS_CACHE_PATH.joinpath(
        shasum(rss_link+'v2')+'.png'
    ))
    if not isfile(favicon_path):
        if image_url:
            try:
                get_favicon(image_url, favicon_path, direct=True)
            except Exception:
                print('Invalid image url for feed `{0}` ({1})'.format(
                    rss_link, image_url
                ))
                image_url = None
        if not image_url:
            try:
                get_favicon(rss_link, favicon_path)
                if not isfile(favicon_path):
                    get_favicon(
                        link or raw_entries[0].uri,
                        favicon_path
                    )
            except Exception:
                print(f'No favicon for feed `{rss_link}`')
                favicon_path = None
    return FeedParserRes(
        is_null=False, error=None, sd_feed=sd_feed,
        rss_link=rss_link,
        title=title,
        link=link,
        description=description,
        image_url=image_url,
        favicon_path=favicon_path,
        raw_entries=raw_entries
    )



from gi.repository import Gtk


def show_shortcuts_window(parent_win, *args):
    shortcuts_win = Gtk.Builder.new_from_resource(
        '/org/gabmus/gfeeds/ui/shortcutsWindow.ui'
    ).get_object('shortcuts-gfeeds')
    shortcuts_win.set_transient_for(parent_win)
    shortcuts_win.present()



from typing import Callable, Optional, List
from gi.repository import Gtk, Adw


class ScrolledDialogResponse:
    def __init__(
            self, name: str, label: str,
            callback: Optional[Callable] = None,
            appearance: Optional[Adw.ResponseAppearance] = None,
    ):
        self.name = name
        self.label = label
        self.callback = callback
        self.appearance = appearance


class ScrolledDialog(Adw.MessageDialog):
    def __init__(
            self, parent: Gtk.Window, title: str, body: str,
            responses: List[ScrolledDialogResponse]
    ):
        self.__parent = parent
        self.__title = title
        self.__body = body
        self.__responses = responses

        super().__init__(
            transient_for=self.__parent,
            heading=self.__title,
            extra_child=Gtk.ScrolledWindow(
                css_classes=['card'],
                hscrollbar_policy=Gtk.PolicyType.NEVER,
                width_request=270,
                height_request=270,
                margin_start=12,
                margin_end=12,
                child=Gtk.Label(
                    wrap=True,
                    xalign=0.0,
                    margin_top=12,
                    margin_bottom=12,
                    margin_start=12,
                    margin_end=12,
                    label=self.__body
                )
            )
        )

        for r in self.__responses:
            self.add_response(r.name, r.label)
            if r.appearance is not None:
                self.set_response_appearance(r.name, r.appearance)

        self.connect('response', self.on_response)

    def on_response(self, dialog: 'ScrolledDialog', res: str):
        for r in self.__responses:
            if r.name == res:
                if r.callback is not None:
                    r.callback(dialog, res)
                return



from typing import Literal, Tuple, Type
from gi.repository import GObject

SignalReturnType = (
    Literal[
        GObject.TYPE_INT,
        GObject.TYPE_BOOLEAN,
        GObject.TYPE_VARIANT,
        GObject.TYPE_FLOAT,
        GObject.TYPE_DOUBLE,
        GObject.TYPE_STRING,
        GObject.TYPE_NONE,
        GObject.TYPE_PYOBJECT,
    ]
    | None
)


def signal_tuple(
    flag: GObject.SignalFlags = GObject.SignalFlags.RUN_LAST,
    ret: SignalReturnType = None,
    params: Tuple[Type] = tuple(),
) -> Tuple[GObject.SignalFlags, SignalReturnType, Tuple[Type]]:
    return (flag, ret, params)



from gettext import gettext as _
from typing import Dict, Optional
from gi.repository import Adw, GObject, Gtk
from gfeeds.confManager import ConfManager
from gfeeds.feeds_manager import FeedsManager
from xml.sax.saxutils import escape
from gfeeds.scrolled_dialog import ScrolledDialogResponse, ScrolledDialog
from gfeeds.webview import GFeedsWebView


VIEW_MODE_ICONS = {
    'webview': 'globe-alt-symbolic',
    'reader': 'ephy-reader-mode-symbolic',
    'feedcont': 'application-rss+xml-symbolic'
}


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/add_feed_popover.ui')
class AddFeedPopover(Gtk.Popover):
    __gtype_name__ = 'AddFeedPopover'
    confirm_btn = Gtk.Template.Child()
    url_entry = Gtk.Template.Child()
    already_subscribed_revealer = Gtk.Template.Child()

    def __init__(self, relative_to):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = FeedsManager()

        relative_to.set_popover(self)

    @Gtk.Template.Callback()
    def on_url_entry_changed(self, *__):
        self.already_subscribed_revealer.set_reveal_child(False)

    @Gtk.Template.Callback()
    def on_url_entry_activate(self, *__):
        if self.confirm_btn.get_sensitive():
            self.on_confirm_btn_clicked(self.confirm_btn)

    @Gtk.Template.Callback()
    def on_confirm_btn_clicked(self, __):
        url = self.url_entry.get_text().strip()
        if not url:
            return
        feed_is_new = self.feedman.add_feed(url, True)
        if feed_is_new:
            self.popdown()
            self.already_subscribed_revealer.set_reveal_child(False)
        else:
            self.already_subscribed_revealer.set_reveal_child(True)


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/right_headerbar.ui')
class RightHeaderbar(Gtk.WindowHandle):
    __gtype_name__ = 'RightHeaderbar'
    __gsignals__ = {
        'go_back': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (str,)
        )
    }
    right_headerbar = Gtk.Template.Child()
    view_mode_menu_btn = Gtk.Template.Child()
    extra_menu_btn = Gtk.Template.Child()
    zoom_in_btn = Gtk.Template.Child()
    zoom_out_btn = Gtk.Template.Child()
    zoom_reset_btn = Gtk.Template.Child()
    title_squeezer = Gtk.Template.Child()
    right_title_container = Gtk.Template.Child()
    title_label = Gtk.Template.Child()
    back_btn = Gtk.Template.Child()

    def __init__(self, webview: Optional[GFeedsWebView] = None):
        super().__init__()
        self.confman = ConfManager()
        self.webview = webview
        self.set_view_mode_icon(self.confman.nconf.default_view)
        self.on_zoom_changed(None, self.confman.nconf.webview_zoom)

    @GObject.Property(type=GFeedsWebView, default=None, nick='webview')
    def webview(self) -> GFeedsWebView:  # type: ignore
        return self.__webview

    @webview.setter
    def webview(self, wv: GFeedsWebView):
        self.__webview = wv
        if wv is None:
            return
        self.__webview.connect('gfeeds_webview_load_start', self.on_load_start)
        self.__webview.connect('zoom_changed', self.on_zoom_changed)

    @Gtk.Template.Callback()
    def on_zoom_in_btn_clicked(self, *__):
        if self.webview is None:
            return
        self.webview.key_zoom_in()

    @Gtk.Template.Callback()
    def on_zoom_out_btn_clicked(self, *__):
        if self.webview is None:
            return
        self.webview.key_zoom_out()

    @Gtk.Template.Callback()
    def on_zoom_reset_btn_clicked(self, *__):
        if self.webview is None:
            return
        self.webview.key_zoom_reset()

    @Gtk.Template.Callback()
    def on_back_btn_clicked(self, *__):
        self.emit('go_back', '')

    def on_zoom_changed(self, __, n_zoom: float):
        self.zoom_reset_btn.set_label(f'{round(n_zoom*100)}%')

    def set_view_mode_icon(self, mode):
        self.view_mode_menu_btn.set_icon_name(
            VIEW_MODE_ICONS[mode]
        )

    def on_view_mode_change(self, target):
        self.set_view_mode_icon(target)

    def set_article_title(self, title):
        self.title_label.set_text(title)

    def on_load_start(self, *_):
        self.view_mode_menu_btn.set_sensitive(True)


@Gtk.Template(resource_path='/org/gabmus/gfeeds/ui/left_headerbar.ui')
class LeftHeaderbar(Gtk.WindowHandle):
    __gtype_name__ = 'LeftHeaderbar'
    left_headerbar = Gtk.Template.Child()
    menu_btn = Gtk.Template.Child()
    filter_btn = Gtk.Template.Child()
    add_btn = Gtk.Template.Child()
    refresh_btn = Gtk.Template.Child()
    search_btn = Gtk.Template.Child()
    errors_btn = Gtk.Template.Child()

    def __init__(self, searchbar: Optional[Gtk.SearchBar] = None):
        super().__init__()
        self.confman = ConfManager()
        self.feedman = FeedsManager()
        self.searchbar = searchbar

        self.add_popover = AddFeedPopover(self.add_btn)
        self.add_btn.set_popover(self.add_popover)

        self.feedman.connect(
            'feedmanager_refresh_start',
            self.on_new_feed_add_start
        )
        self.feedman.connect(
            'feedmanager_refresh_end',
            self.on_new_feed_add_end
        )
        self.on_new_feed_add_start()

    @GObject.Property(
        type=Gtk.SearchBar, default=None, nick='searchbar'
    )
    def searchbar(self) -> Gtk.SearchBar:  # type: ignore
        return self.__searchbar

    @searchbar.setter
    def searchbar(self, sb: Gtk.SearchBar):
        self.__searchbar = sb

    @Gtk.Template.Callback()
    def on_refresh_btn_clicked(self, *__):
        self.feedman.refresh()

    @Gtk.Template.Callback()
    def show_errors_dialog(self, *__):

        def on_remove(d: ScrolledDialog, _):
            d.close()
            feeds: Dict[str, dict] = self.confman.nconf.feeds
            for pf in self.feedman.problematic_feeds:
                if pf in feeds.keys():
                    feeds.pop(pf)
            self.confman.nconf.feeds = feeds
            self.errors_btn.set_visible(False)

        def on_keep(d: ScrolledDialog, _):
            d.close()
            self.errors_btn.set_visible(True)

        dialog = ScrolledDialog(
            self.get_root(),  # type: ignore
            _(
                'There were problems with some feeds.'
                ' Do you want to remove them?'
            ),
            escape('\n'.join(self.feedman.errors)),
            [
                ScrolledDialogResponse('keep', _('_Keep'), on_keep),
                ScrolledDialogResponse(
                    'remove', _('_Remove'), on_remove,
                    Adw.ResponseAppearance.DESTRUCTIVE
                )
            ]
        )
        dialog.present()

    def on_new_feed_add_start(self, *_):
        self.refresh_btn.set_sensitive(False)
        self.add_popover.confirm_btn.set_sensitive(False)

    def on_new_feed_add_end(self, *_):
        self.refresh_btn.set_sensitive(True)
        self.add_popover.confirm_btn.set_sensitive(True)
        self.add_popover.url_entry.set_text('')
